function D = sparseFD(k,m,h,N)
warning off;
% D is the sparse FD matrix computing the m^th order differentiation
% the grid is required to be equispaced, with the interval being "h"
% periodic boundary condition
% "m" specifies the desired order of differentiation, and m < k
% "k" specifies the number of points used, here we require k is odd;
% 'N' specifies the dimensioin of the matrix D

if rem(k,2) == 0
    error('k must be an odd number!!!');
end
if ((m >= k) || (N < k))
    error('Must: m < k <= N !!!');
end

%% The main part of the Differentiation matrix: central FD
em = [zeros(m,1); 1; zeros(k-m-1,1)];

for ik1 = 1:k % counting the rows
    for ik2 = 1:k % counting the columns
        M(ik1,ik2) = ((ik2-(k+1)/2)*h)^(ik1-1)/factorial(ik1-1);
    end    
end
c = M\em;
c = [c((k+1)/2+1:end); c; c(1:(k+1)/2-1)];
C = ones(N,1)*c.';

d = [(1-N):1:((k-1)/2-N) -(k-1)/2:(k-1)/2 (N-(k-1)/2):1:(N-1)];
D = spdiags(C,d,N,N);
end